﻿namespace HeuristicLab.Problems.ProgramSynthesis.Push.Expressions {
  using System;
  using Common;

  using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;
  using HeuristicLab.Problems.ProgramSynthesis.Push.Attributes;
  using HeuristicLab.Problems.ProgramSynthesis.Push.Interpreter;
  using HeuristicLab.Problems.ProgramSynthesis.Push.Stack;

  /// <summary>
  ///     Pushes the sum of the top two items.
  /// </summary>
  [PushExpression(StackTypes.Float, "FLOAT.+")]
  [StorableClass]
  public class FloatAddExpression : StatelessExpression {
    public FloatAddExpression() { }
    [StorableConstructor]
    protected FloatAddExpression(bool deserializing) : base(deserializing) { }

    public override bool IsNoop(IInternalPushInterpreter interpreter) {
      return interpreter.FloatStack.Count < 2;
    }

    public override void Eval(IInternalPushInterpreter interpreter) {
      var first = interpreter.FloatStack.Pop();
      var second = interpreter.FloatStack.Top;
      var result = second + first;

      if (double.IsPositiveInfinity(result)) result = double.MaxValue;
      if (double.IsNegativeInfinity(result)) result = double.MinValue;

      interpreter.FloatStack.SetTop(result);
    }
  }

  /// <summary>
  ///     Pushes the difference of the top two items; that is, the second item minus the top item.
  /// </summary>
  [PushExpression(StackTypes.Float, "FLOAT.-")]
  [StorableClass]
  public class FloatSubtractExpression : StatelessExpression {
    public FloatSubtractExpression() { }
    [StorableConstructor]
    protected FloatSubtractExpression(bool deserializing) : base(deserializing) { }

    public override bool IsNoop(IInternalPushInterpreter interpreter) {
      return interpreter.FloatStack.Count < 2;
    }

    public override void Eval(IInternalPushInterpreter interpreter) {
      var first = interpreter.FloatStack.Pop();
      var second = interpreter.FloatStack.Top;
      var result = second - first;

      if (double.IsPositiveInfinity(result)) result = double.MaxValue;
      if (double.IsNegativeInfinity(result)) result = double.MinValue;

      interpreter.FloatStack.SetTop(result);
    }
  }

  /// <summary>
  ///     Pushes the product of the top two items.
  /// </summary>
  [PushExpression(StackTypes.Float, "FLOAT.*")]
  [StorableClass]
  public class FloatMultiplyExpression : StatelessExpression {
    public FloatMultiplyExpression() { }
    [StorableConstructor]
    protected FloatMultiplyExpression(bool deserializing) : base(deserializing) { }

    public override bool IsNoop(IInternalPushInterpreter interpreter) {
      return interpreter.FloatStack.Count < 2;
    }

    public override void Eval(IInternalPushInterpreter interpreter) {
      var first = interpreter.FloatStack.Pop();
      var second = interpreter.FloatStack.Top;
      var result = second * first;

      if (double.IsPositiveInfinity(result)) result = double.MaxValue;
      if (double.IsNegativeInfinity(result)) result = double.MinValue;

      interpreter.FloatStack.SetTop(result);
    }
  }

  /// <summary>
  ///     Pushes the quotient of the top two items; that is, the second item divided by the top item.
  ///     If the top item is zero this acts as a NOOP.
  /// </summary>
  [PushExpression(StackTypes.Float, "FLOAT./")]
  [StorableClass]
  public class FloatDivideExpression : StatelessExpression {
    public FloatDivideExpression() { }
    [StorableConstructor]
    protected FloatDivideExpression(bool deserializing) : base(deserializing) { }

    public override bool IsNoop(IInternalPushInterpreter interpreter) {
      return interpreter.FloatStack.Count < 2 || interpreter.FloatStack.Top.IsAlmost(0);
    }

    public override void Eval(IInternalPushInterpreter interpreter) {
      var first = interpreter.FloatStack.Pop();
      var second = interpreter.FloatStack.Top;
      var result = second / first;

      if (double.IsPositiveInfinity(result)) result = double.MaxValue;
      if (double.IsNegativeInfinity(result)) result = double.MinValue;

      interpreter.FloatStack.SetTop(result);
    }
  }

  /// <summary>
  ///     Pushes the second stack item modulo the top stack item. If the top item is zero this acts as a NOOP.
  ///     The modulus is computed as the remainder of the quotient, where the quotient has first been truncated toward
  ///     negative infinity.
  ///     (This is taken from the definition for the generic MOD function in Common Lisp, which is described
  ///     for example at http://www.lispworks.com/reference/HyperSpec/Body/f_mod_r.htm.)
  /// </summary>
  [PushExpression(StackTypes.Float, "FLOAT.%")]
  [StorableClass]
  public class FloatModuloExpression : StatelessExpression {
    public FloatModuloExpression() { }
    [StorableConstructor]
    protected FloatModuloExpression(bool deserializing) : base(deserializing) { }

    public override bool IsNoop(IInternalPushInterpreter interpreter) {
      return interpreter.FloatStack.Count < 2 || interpreter.FloatStack.Top.IsAlmost(0);
    }

    public override void Eval(IInternalPushInterpreter interpreter) {
      var first = interpreter.FloatStack.Pop();
      var second = interpreter.FloatStack.Top;
      var result = second % first;

      interpreter.FloatStack.SetTop(result);
    }
  }

  /// <summary>
  ///     Pushes the minimum of the top two items.
  /// </summary>
  [PushExpression(StackTypes.Float, "FLOAT.MIN")]
  [StorableClass]
  public class FloatMinExpression : StatelessExpression {
    public FloatMinExpression() { }
    [StorableConstructor]
    protected FloatMinExpression(bool deserializing) : base(deserializing) { }

    public override bool IsNoop(IInternalPushInterpreter interpreter) {
      return interpreter.FloatStack.Count < 2;
    }

    public override void Eval(IInternalPushInterpreter interpreter) {
      var first = interpreter.FloatStack.Pop();
      var second = interpreter.FloatStack.Top;
      var result = Math.Min(second, first);

      interpreter.FloatStack.SetTop(result);
    }
  }

  /// <summary>
  ///     Pushes the maximum of the top two items.
  /// </summary>
  [PushExpression(StackTypes.Float, "FLOAT.MAX")]
  [StorableClass]
  public class FloatMaxExpression : StatelessExpression {
    public FloatMaxExpression() { }
    [StorableConstructor]
    protected FloatMaxExpression(bool deserializing) : base(deserializing) { }

    public override bool IsNoop(IInternalPushInterpreter interpreter) {
      return interpreter.FloatStack.Count < 2;
    }

    public override void Eval(IInternalPushInterpreter interpreter) {
      var first = interpreter.FloatStack.Pop();
      var second = interpreter.FloatStack.Top;
      var result = Math.Max(second, first);

      interpreter.FloatStack.SetTop(result);
    }
  }

  /// <summary>
  ///     Pushes TRUE onto the BOOLEAN stack if the second item is less than the top item, or FALSE otherwise.
  /// </summary>
  [PushExpression(StackTypes.Float, "FLOAT.<", StackTypes.Boolean)]
  [StorableClass]
  public class FloatSmallerThanExpression : StatelessExpression {
    public FloatSmallerThanExpression() { }
    [StorableConstructor]
    protected FloatSmallerThanExpression(bool deserializing) : base(deserializing) { }

    public override bool IsNoop(IInternalPushInterpreter interpreter) {
      return interpreter.FloatStack.Count < 2;
    }

    public override void Eval(IInternalPushInterpreter interpreter) {
      var first = interpreter.FloatStack.Top;
      var second = interpreter.FloatStack[1];
      interpreter.FloatStack.Remove(2);

      var result = second < first;
      interpreter.BooleanStack.Push(result);
    }
  }

  /// <summary>
  ///     Pushes TRUE onto the BOOLEAN stack if the second item is less than the top item, or FALSE otherwise.
  /// </summary>
  [PushExpression(StackTypes.Float, "FLOAT.<=", StackTypes.Boolean)]
  [StorableClass]
  public class FloatSmallerThanOrEqualExpression : StatelessExpression {
    public FloatSmallerThanOrEqualExpression() { }
    [StorableConstructor]
    protected FloatSmallerThanOrEqualExpression(bool deserializing) : base(deserializing) { }

    public override bool IsNoop(IInternalPushInterpreter interpreter) {
      return interpreter.FloatStack.Count < 2;
    }

    public override void Eval(IInternalPushInterpreter interpreter) {
      var first = interpreter.FloatStack.Top;
      var second = interpreter.FloatStack[1];
      interpreter.FloatStack.Remove(2);

      var result = second < first || second.IsAlmost(first);
      interpreter.BooleanStack.Push(result);
    }
  }

  /// <summary>
  ///     Pushes TRUE onto the BOOLEAN stack if the second item is greater than the top item, or FALSE otherwise.
  /// </summary>
  [PushExpression(StackTypes.Float, "FLOAT.>", StackTypes.Boolean)]
  [StorableClass]
  public class FloatGreaterThanExpression : StatelessExpression {
    public FloatGreaterThanExpression() { }
    [StorableConstructor]
    protected FloatGreaterThanExpression(bool deserializing) : base(deserializing) { }

    public override bool IsNoop(IInternalPushInterpreter interpreter) {
      return interpreter.FloatStack.Count < 2;
    }

    public override void Eval(IInternalPushInterpreter interpreter) {
      var first = interpreter.FloatStack.Top;
      var second = interpreter.FloatStack[1];
      interpreter.FloatStack.Remove(2);

      var result = second > first;
      interpreter.BooleanStack.Push(result);
    }
  }

  /// <summary>
  ///     Pushes TRUE onto the BOOLEAN stack if the second item is greater than or equal to the top item, or FALSE otherwise.
  /// </summary>
  [PushExpression(StackTypes.Float, "FLOAT.>=", StackTypes.Boolean)]
  [StorableClass]
  public class FloatGreaterThanOrEqualExpression : StatelessExpression {
    public FloatGreaterThanOrEqualExpression() { }
    [StorableConstructor]
    protected FloatGreaterThanOrEqualExpression(bool deserializing) : base(deserializing) { }

    public override bool IsNoop(IInternalPushInterpreter interpreter) {
      return interpreter.FloatStack.Count < 2;
    }

    public override void Eval(IInternalPushInterpreter interpreter) {
      var first = interpreter.FloatStack.Top;
      var second = interpreter.FloatStack[1];
      interpreter.FloatStack.Remove(2);

      var result = second > first || second.IsAlmost(first);
      interpreter.BooleanStack.Push(result);
    }
  }

  /// <summary>
  ///     Pushes the sine of the top item.
  /// </summary>
  [PushExpression(StackTypes.Float, "FLOAT.SIN")]
  [StorableClass]
  public class FloatSineExpression : StatelessExpression {
    public FloatSineExpression() { }
    [StorableConstructor]
    protected FloatSineExpression(bool deserializing) : base(deserializing) { }

    public override bool IsNoop(IInternalPushInterpreter interpreter) {
      return interpreter.FloatStack.IsEmpty;
    }

    public override void Eval(IInternalPushInterpreter interpreter) {
      var result = Math.Sin(interpreter.FloatStack.Top);
      interpreter.FloatStack.SetTop(result);
    }
  }

  /// <summary>
  ///     Pushes the cosine of the top item.
  /// </summary>
  [PushExpression(StackTypes.Float, "FLOAT.COS")]
  public class FloatCosineExpression : StatelessExpression {
    public FloatCosineExpression() { }
    [StorableConstructor]
    protected FloatCosineExpression(bool deserializing) : base(deserializing) { }

    public override bool IsNoop(IInternalPushInterpreter interpreter) {
      return interpreter.FloatStack.IsEmpty;
    }

    public override void Eval(IInternalPushInterpreter interpreter) {
      var result = Math.Cos(interpreter.FloatStack.Top);
      interpreter.FloatStack.SetTop(result);
    }
  }

  /// <summary>
  ///     Pushes 1 if the top BOOLEAN is TRUE, or 0 if the top BOOLEAN is FALSE.
  /// </summary>
  [PushExpression(StackTypes.Float, "FLOAT.FROMBOOLEAN", StackTypes.Boolean)]
  [StorableClass]
  public class FloatFromBooleanExpression : StatelessExpression {
    public FloatFromBooleanExpression() { }
    [StorableConstructor]
    protected FloatFromBooleanExpression(bool deserializing) : base(deserializing) { }

    public override bool IsNoop(IInternalPushInterpreter interpreter) {
      return interpreter.BooleanStack.IsEmpty;
    }

    public override void Eval(IInternalPushInterpreter interpreter) {
      var condition = interpreter.BooleanStack.Pop();
      var value = condition ? 1 : 0;

      interpreter.FloatStack.Push(value);
    }
  }

  /// <summary>
  ///     Pushes a floating point version of the top INTEGER.
  /// </summary>
  [PushExpression(StackTypes.Float, "FLOAT.FROMINTEGER", StackTypes.Integer)]
  [StorableClass]
  public class FloatFromIntegerExpression : StatelessExpression {
    public FloatFromIntegerExpression() { }
    [StorableConstructor]
    protected FloatFromIntegerExpression(bool deserializing) : base(deserializing) { }

    public override bool IsNoop(IInternalPushInterpreter interpreter) {
      return interpreter.IntegerStack.IsEmpty;
    }

    public override void Eval(IInternalPushInterpreter interpreter) {
      var value = (double)interpreter.IntegerStack.Pop();
      interpreter.FloatStack.Push(value);
    }
  }
}