﻿namespace HeuristicLab.Problems.ProgramSynthesis.Push.Expressions {
  using System.Collections.Generic;

  using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;
  using HeuristicLab.Problems.ProgramSynthesis.Push.Attributes;
  using HeuristicLab.Problems.ProgramSynthesis.Push.Interpreter;
  using HeuristicLab.Problems.ProgramSynthesis.Push.Stack;

  /// <summary>
  /// Concats two vectors on the type stack.
  /// </summary>
  [StorableClass]
  public abstract class VectorConcatExpression<T> : StatelessExpression {
    protected VectorConcatExpression() { }
    [StorableConstructor]
    protected VectorConcatExpression(bool deserializing) : base(deserializing) { }

    protected bool IsNoop(IInternalPushInterpreter interpreter, IPushStack<List<T>> vectorStack) {
      return vectorStack.Count < 2 ||
             vectorStack.Top.Count + vectorStack[1].Count > interpreter.Configuration.MaxVectorLength;
    }

    protected void Eval(IInternalPushInterpreter interpter, IPushStack<List<T>> vectorStack) {
      var first = vectorStack.Pop();
      var second = vectorStack.Top;
      var result = new T[first.Count + second.Count];

      first.CopyTo(result, 0);
      second.CopyTo(result, first.Count);
      vectorStack.SetTop(new List<T>(result));
    }
  }

  [StorableClass]
  [PushExpression(StackTypes.IntegerVector, "INTEGER[].CONCAT")]
  public class IntegerVectorConcatExpression : VectorConcatExpression<long> {
    public IntegerVectorConcatExpression() { }
    [StorableConstructor]
    protected IntegerVectorConcatExpression(bool deserializing) : base(deserializing) { }

    public override bool IsNoop(IInternalPushInterpreter interpreter) {
      return IsNoop(interpreter, interpreter.IntegerVectorStack);
    }

    public override void Eval(IInternalPushInterpreter interpreter) {
      Eval(interpreter, interpreter.IntegerVectorStack);
    }
  }

  [StorableClass]
  [PushExpression(StackTypes.FloatVector, "FLOAT[].CONCAT")]
  public class FloatVectorConcatExpression : VectorConcatExpression<double> {
    public FloatVectorConcatExpression() { }
    [StorableConstructor]
    protected FloatVectorConcatExpression(bool deserializing) : base(deserializing) { }

    public override bool IsNoop(IInternalPushInterpreter interpreter) {
      return IsNoop(interpreter, interpreter.FloatVectorStack);
    }

    public override void Eval(IInternalPushInterpreter interpreter) {
      Eval(interpreter, interpreter.FloatVectorStack);
    }
  }

  [StorableClass]
  [PushExpression(StackTypes.BooleanVector, "BOOLEAN[].CONCAT")]
  public class BooleanVectorConcatExpression : VectorConcatExpression<bool> {
    public BooleanVectorConcatExpression() { }
    [StorableConstructor]
    protected BooleanVectorConcatExpression(bool deserializing) : base(deserializing) { }

    public override bool IsNoop(IInternalPushInterpreter interpreter) {
      return IsNoop(interpreter, interpreter.BooleanVectorStack);
    }

    public override void Eval(IInternalPushInterpreter interpreter) {
      Eval(interpreter, interpreter.BooleanVectorStack);
    }
  }

  [StorableClass]
  [PushExpression(StackTypes.StringVector, "STRING[].CONCAT")]
  public class StringVectorConcatExpression : VectorConcatExpression<string> {
    public StringVectorConcatExpression() { }
    [StorableConstructor]
    protected StringVectorConcatExpression(bool deserializing) : base(deserializing) { }

    public override bool IsNoop(IInternalPushInterpreter interpreter) {
      return IsNoop(interpreter, interpreter.StringVectorStack);
    }

    public override void Eval(IInternalPushInterpreter interpreter) {
      Eval(interpreter, interpreter.StringVectorStack);
    }
  }
}
