﻿using System.Collections.Generic;

namespace HeuristicLab.Problems.ProgramSynthesis.Push.Expressions {
  using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;
  using HeuristicLab.Problems.ProgramSynthesis.Push.Attributes;
  using HeuristicLab.Problems.ProgramSynthesis.Push.Interpreter;
  using HeuristicLab.Problems.ProgramSynthesis.Push.Stack;

  /// <summary>
  /// Takes the first N items from the type stack, where N is from the integer stack
  /// </summary>
  /// <typeparam name="T"></typeparam>
  [StorableClass]
  public abstract class VectorTakeExpression<T> : StatelessExpression {
    protected VectorTakeExpression() { }
    [StorableConstructor]
    protected VectorTakeExpression(bool deserializing) : base(deserializing) { }

    protected bool IsNoop(IInternalPushInterpreter interpreter, IPushStack<List<T>> vectorStack) {
      return vectorStack.IsEmpty || interpreter.IntegerStack.IsEmpty;
    }

    protected void Eval(IInternalPushInterpreter interpreter, IPushStack<List<T>> vectorStack) {
      var count = (int)interpreter.IntegerStack.Pop();

      if (vectorStack.Top.Count == 0)
        return;

      count %= vectorStack.Top.Count;

      if (count < 0)
        count *= -1;

      var result = vectorStack.Top.GetRange(0, count);
      vectorStack.SetTop(result);
    }
  }

  [StorableClass]
  [PushExpression(StackTypes.IntegerVector, "INTEGER[].TAKE", StackTypes.Integer)]
  public class IntegerVectorTakeExpression : VectorTakeExpression<long> {
    public IntegerVectorTakeExpression() { }
    [StorableConstructor]
    protected IntegerVectorTakeExpression(bool deserializing) : base(deserializing) { }

    public override bool IsNoop(IInternalPushInterpreter interpreter) {
      return IsNoop(interpreter, interpreter.IntegerVectorStack);
    }

    public override void Eval(IInternalPushInterpreter interpreter) {
      Eval(interpreter, interpreter.IntegerVectorStack);
    }
  }

  [StorableClass]
  [PushExpression(StackTypes.FloatVector, "FLOAT[].TAKE", StackTypes.Integer)]
  public class FloatVectorTakeExpression : VectorTakeExpression<double> {
    public FloatVectorTakeExpression() { }
    [StorableConstructor]
    protected FloatVectorTakeExpression(bool deserializing) : base(deserializing) { }

    public override bool IsNoop(IInternalPushInterpreter interpreter) {
      return IsNoop(interpreter, interpreter.FloatVectorStack);
    }

    public override void Eval(IInternalPushInterpreter interpreter) {
      Eval(interpreter, interpreter.FloatVectorStack);
    }
  }

  [StorableClass]
  [PushExpression(StackTypes.BooleanVector, "BOOLEAN[].TAKE", StackTypes.Integer)]
  public class BooleanVectorTakeExpression : VectorTakeExpression<bool> {
    public BooleanVectorTakeExpression() { }
    [StorableConstructor]
    protected BooleanVectorTakeExpression(bool deserializing) : base(deserializing) { }

    public override bool IsNoop(IInternalPushInterpreter interpreter) {
      return IsNoop(interpreter, interpreter.BooleanVectorStack);
    }

    public override void Eval(IInternalPushInterpreter interpreter) {
      Eval(interpreter, interpreter.BooleanVectorStack);
    }
  }

  [StorableClass]
  [PushExpression(StackTypes.StringVector, "STRING[].TAKE", StackTypes.Integer)]
  public class StringVectorTakeExpression : VectorTakeExpression<string> {
    public StringVectorTakeExpression() { }
    [StorableConstructor]
    protected StringVectorTakeExpression(bool deserializing) : base(deserializing) { }

    public override bool IsNoop(IInternalPushInterpreter interpreter) {
      return IsNoop(interpreter, interpreter.StringVectorStack);
    }

    public override void Eval(IInternalPushInterpreter interpreter) {
      Eval(interpreter, interpreter.StringVectorStack);
    }
  }
}
